#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2025 The Electrum Developers
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio
import concurrent
from typing import TYPE_CHECKING, List, Tuple, Optional

from PyQt6.QtCore import QObject, pyqtSignal, pyqtProperty, pyqtSlot

from electrum import util
from electrum.plugin import hook
from electrum.transaction import PartialTransaction, tx_from_any
from electrum.wallet import Multisig_Wallet
from electrum.util import EventListener, event_listener

from electrum.gui.qml.qewallet import QEWallet

from .psbt_nostr import PsbtNostrPlugin, CosignerWallet

if TYPE_CHECKING:
    from electrum.wallet import Abstract_Wallet
    from electrum.gui.qml import ElectrumQmlApplication


class QReceiveSignalObject(QObject):
    def __init__(self, plugin: 'Plugin'):
        QObject.__init__(self)
        self._plugin = plugin

    cosignerReceivedPsbt = pyqtSignal(str, str, str, str, bool)
    sendPsbtFailed = pyqtSignal(str, arguments=['reason'])
    sendPsbtSuccess = pyqtSignal()

    @pyqtProperty(str)
    def loader(self):
        return 'main.qml'

    @pyqtSlot(QEWallet, str, result=bool)
    def canSendPsbt(self, wallet: 'QEWallet', tx: str) -> bool:
        cosigner_wallet = self._plugin.cosigner_wallets.get(wallet.wallet)
        if not cosigner_wallet:
            return False
        return cosigner_wallet.can_send_psbt(tx_from_any(tx, deserialize=True))

    @pyqtSlot(QEWallet, str)
    @pyqtSlot(QEWallet, str, str)
    def sendPsbt(self, wallet: 'QEWallet', tx: str, label: str = None):
        cosigner_wallet = self._plugin.cosigner_wallets.get(wallet.wallet)
        if not cosigner_wallet:
            return
        cosigner_wallet.send_psbt(tx_from_any(tx, deserialize=True), label)

    @pyqtSlot(QEWallet, str, str)
    def saveTxLabel(self, wallet: 'QEWallet', tx: str, label: str):
        cosigner_wallet = self._plugin.cosigner_wallets.get(wallet.wallet)
        if not cosigner_wallet:
            return
        cosigner_wallet.save_tx_label(tx_from_any(tx, deserialize=True), label)

    @pyqtSlot(QEWallet, str)
    @pyqtSlot(QEWallet, str, bool)
    def acceptPsbt(self, wallet: 'QEWallet', event_id: str, save_to_wallet: bool = False):
        cosigner_wallet = self._plugin.cosigner_wallets.get(wallet.wallet)
        if not cosigner_wallet:
            return
        cosigner_wallet.accept_psbt(event_id, save_to_wallet)
        if save_to_wallet:
            # let GUI update view through wallet_updated callback
            util.trigger_callback('wallet_updated', wallet.wallet)

    @pyqtSlot(QEWallet, str)
    def rejectPsbt(self, wallet: 'QEWallet', event_id: str):
        cosigner_wallet = self._plugin.cosigner_wallets.get(wallet.wallet)
        if not cosigner_wallet:
            return
        cosigner_wallet.reject_psbt(event_id)


class Plugin(PsbtNostrPlugin):
    def __init__(self, parent, config, name):
        super().__init__(parent, config, name)
        self.so = QReceiveSignalObject(self)
        self._app = None

    @hook
    def init_qml(self, app: 'ElectrumQmlApplication'):
        self._app = app
        self.so.setParent(app)  # parent in QObject tree
        # plugin enable for already open wallet
        wallet = app.daemon.currentWallet.wallet if app.daemon.currentWallet else None
        if wallet:
            self.load_wallet(wallet)

    @hook
    def load_wallet(self, wallet: 'Abstract_Wallet'):
        # remove existing, only foreground wallet active
        for _wallet in self.cosigner_wallets.copy().keys():
            self.remove_cosigner_wallet(_wallet)
        if not isinstance(wallet, Multisig_Wallet):
            return
        if wallet.wallet_type == '2fa':
            return
        self.add_cosigner_wallet(wallet, QmlCosignerWallet(wallet, self))


class QmlCosignerWallet(EventListener, CosignerWallet):

    def __init__(self, wallet: 'Multisig_Wallet', plugin: 'Plugin'):
        db_storage = plugin.get_storage(wallet)
        CosignerWallet.__init__(self, wallet, db_storage)
        self.plugin = plugin
        self.register_callbacks()

        self.tx = None

    @event_listener
    def on_event_psbt_nostr_received(self, wallet, pubkey, event_id, tx: 'PartialTransaction', label: str):
        if self.wallet == wallet:
            self.tx = tx
            can_be_saved = tx.txid() is not None
            self.plugin.so.cosignerReceivedPsbt.emit(pubkey, event_id, tx.serialize(), label, can_be_saved)

    def close(self):
        super().close()
        self.unregister_callbacks()

    def do_send(self, messages: List[Tuple[str, dict]], txid: Optional[str] = None):
        if not messages:
            return
        coro = self.send_direct_messages(messages)

        loop = util.get_asyncio_loop()
        assert util.get_running_loop() != loop, 'must not be called from asyncio thread'
        self._result = None
        self._future = asyncio.run_coroutine_threadsafe(coro, loop)

        try:
            self._result = self._future.result()
            self.plugin.so.sendPsbtSuccess.emit()
        except concurrent.futures.CancelledError:
            pass
        except Exception as e:
            self.plugin.so.sendPsbtFailed.emit(str(e))

    def save_tx_label(self, tx, label):
        self.wallet.set_label(tx.txid(), label)

    def accept_psbt(self, event_id, save: bool = False):
        if save:
            self.add_transaction_to_wallet(self.tx, on_failure=self.on_add_fail)
        self.mark_pending_event_rcvd(event_id)

    def reject_psbt(self, event_id):
        self.mark_pending_event_rcvd(event_id)

    def on_add_fail(self, error_msg: str):
        self.logger.error(f'failed to add tx to wallet: {error_msg}')



#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2012 thomasv@gitorious
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import functools
import json
import os
import string
from typing import Optional

import gettext

from .logging import get_logger


_logger = get_logger(__name__)
LOCALE_DIR = os.path.join(os.path.dirname(__file__), 'locale', 'locale')


def _get_null_translations():
    """Returns a gettext Translations obj with translations explicitly disabled."""
    return gettext.translation('electrum', fallback=True, class_=gettext.NullTranslations)


# Set initial default language to None. i.e. translations explicitly disabled.
# The main script or GUIs can call set_language to enable translations.
_language = _get_null_translations()


def _ensure_translation_keeps_format_string_syntax_similar(translator):
    """This checks that the source string is syntactically similar to the translated string.
    If not, translations are rejected by falling back to the source string.
    """
    sf = string.Formatter()
    @functools.wraps(translator)
    def safe_translator(msg: str, **kwargs):
        translation = translator(msg, **kwargs)
        parsed1 = list(sf.parse(msg))  # iterable of tuples (literal_text, field_name, format_spec, conversion)
        try:
            parsed2 = list(sf.parse(translation))
        except ValueError:  # malformed format string in translation
            _logger.warning(
                f"rejected translation string: failed to parse. original={msg!r}. {translation=!r}",
                only_once=True)
            return msg
        # num of replacement fields must match:
        if len(parsed1) != len(parsed2):
            _logger.warning(
                f"rejected translation string: num replacement fields mismatch. original={msg!r}. {translation=!r}",
                only_once=True)
            return msg
        # set of "field_name"s must not change. (re-ordering is explicitly allowed):
        field_names1 = set(tupl[1] for tupl in parsed1)
        field_names2 = set(tupl[1] for tupl in parsed2)
        if field_names1 != field_names2:
            _logger.warning(
                f"rejected translation string: set of field_names mismatch. original={msg!r}. {translation=!r}",
                only_once=True)
            return msg
        # checks done.
        return translation
    return safe_translator


# note: do not use old-style (%) formatting inside translations,
#       as syntactically incorrectly translated strings often raise exceptions (see #3237).
#       e.g. consider  _("Connected to %d nodes.") % n            # <- raises. do NOT use
#                      >>> "Connecté aux noeuds" % n
#                      TypeError: not all arguments converted during string formatting
# note: f-strings cannot be translated! see https://stackoverflow.com/q/49797658
#       So this does NOT work:   _(f"My name: {name}")            # <- cannot be translated. do NOT use
#       instead use .format:     _("My name: {}").format(name)    # <- works. prefer this way.
# note: positional and keyword-based substitution also works with str.format().
#       These give more flexibility to translators: it allows reordering the substituted values.
#       However, only if the translators understand and use it correctly!
#          _("time left: {0} minutes, {1} seconds").format(t//60, t%60)                   # <- works. ok to use
#          _("time left: {mins} minutes, {secs} seconds").format(mins=t//60, secs=t%60)   # <- works, but too complex
@_ensure_translation_keeps_format_string_syntax_similar
def _(msg: str, *, context=None) -> str:
    if msg == "":
        return ""  # empty string must not be translated. see #7158
    if context:
        contexts = [context]
        if context[-1] != "|":  # try with both "|" suffix and without
            contexts.append(context + "|")
        else:
            contexts.append(context[:-1])
        for ctx in contexts:
            out = _language.pgettext(ctx, msg)
            if out != msg:  # found non-trivial translation
                return out
        # else try without context
    return _language.gettext(msg)


def set_language(x: Optional[str]) -> None:
    _logger.info(f"setting language to {x!r}")
    global _language
    if not x:
        return
    if x.startswith("en_"):
        # Setting the language to "English" is a protected special-case:
        # we disable all translations and use the source strings.
        _language = _get_null_translations()
    else:
        _language = gettext.translation('electrum', LOCALE_DIR, fallback=True, languages=[x])


# note: The values (human-visible lang names) should be either in English or in their own lang,
#       but NOT translated to the currently selected lang.
#       e.g. "fr_FR" we could show as either "French" or "Francais", or even as "French - Francais",
#       but it is evil to show it as "Franzosisch". How am I supposed to switch back to English from Korean??? :)
languages = {
    '': _('Default'),
    'ar_SA': 'Arabic',
    'bg_BG': 'Bulgarian',
    'cs_CZ': 'Czech',
    'da_DK': 'Danish',
    'de_DE': 'German',
    'el_GR': 'Greek',
    'eo_UY': 'Esperanto',
    'en_UK': 'English',  # selecting this guarantees seeing the untranslated source strings
    'es_ES': 'Spanish',
    'fa_IR': 'Persian',
    'fr_FR': 'French',
    'hu_HU': 'Hungarian',
    'hy_AM': 'Armenian',
    'id_ID': 'Indonesian',
    'it_IT': 'Italian',
    'ja_JP': 'Japanese',
    'ky_KG': 'Kyrgyz',
    'lv_LV': 'Latvian',
    'nb_NO': 'Norwegian Bokmal',
    'nl_NL': 'Dutch',
    'pl_PL': 'Polish',
    'pt_BR': 'Portuguese (Brazil)',
    'pt_PT': 'Portuguese',
    'ro_RO': 'Romanian',
    'ru_RU': 'Russian',
    'sk_SK': 'Slovak',
    'sl_SI': 'Slovenian',
    'sv_SE': 'Swedish',
    'ta_IN': 'Tamil',
    'th_TH': 'Thai',
    'tr_TR': 'Turkish',
    'uk_UA': 'Ukrainian',
    'vi_VN': 'Vietnamese',
    'zh_CN': 'Chinese Simplified',
    'zh_TW': 'Chinese Traditional',
}
assert '' in languages


def get_gui_lang_names(*, show_completion_percent: bool = True) -> dict[str, str]:
    """Returns a  lang_code -> lang_name  mapping, sorted.

    If show_completion_percent is True, lang_name includes a % estimate for translation completeness.
    """
    # calc catalog sizes
    if show_completion_percent:
        stats = _get_stats()
    # sort ("Default" first, then "English", then lexicographically sorted names)
    languages_copy = languages.copy()
    lang_pair_default = ("", languages_copy.pop("")) # pop "Default"
    lang_pair_english = ("en_UK", languages_copy.pop("en_UK")) # pop "English"
    lang_pairs_sorted = sorted(languages_copy.items(), key=lambda x: x[1])
    # fancy names
    gui_lang_names = {}  # type: dict[str, str]
    gui_lang_names[lang_pair_default[0]] = lang_pair_default[1]
    gui_lang_names[lang_pair_english[0]] = lang_pair_english[1]
    for lang_code, lang_name in lang_pairs_sorted:
        if show_completion_percent and stats:
            source_str_cnt = max(stats["source_string_count"], 1)  # avoid div-by-zero
            try:
                lang_data = stats["translations"][lang_code]
            except KeyError as e:
                _logger.warning(f"missing language from stats.json: {e!r}")
                catalog_percent = "??"
            else:
                translated_str_cnt = lang_data["string_count"]
                catalog_percent = round(100 * translated_str_cnt / source_str_cnt)
            gui_lang_names[lang_code] = f"{lang_name} ({catalog_percent}%)"
        else:
            gui_lang_names[lang_code] = lang_name
    return gui_lang_names


_stats = None
def _get_stats() -> dict:
    global _stats
    if _stats is None:
        fname = f"{LOCALE_DIR}/stats.json"
        try:
            with open(fname, "r", encoding="utf-8") as f:
                text = f.read()
        except OSError as e:  # we tolerate the file missing
            # This can happen e.g. when running from git clone if user did not run build_locale.sh.
            _logger.info(f"failed to open stats file {fname!r} - built locale (translations) missing??: {e!r}")
            _stats = {}
        else:  # found file. if it is there, it MUST parse correctly
            _stats = json.loads(text)
    return _stats



# Copyright (C) 2018 The Electrum developers
# Distributed under the MIT software license, see the accompanying
# file LICENCE or http://www.opensource.org/licenses/mit-license.php

import asyncio
from typing import TYPE_CHECKING, Optional, Dict, Callable, Awaitable

from . import util
from .util import (
    TxMinedInfo, BelowDustLimit, NoDynamicFeeEstimates, OldTaskGroup, EventListener, event_listener, log_exceptions,
    ignore_exceptions, now
)
from .transaction import Transaction, TxOutpoint
from .logging import Logger
from .address_synchronizer import TX_HEIGHT_LOCAL
from .lnutil import REDEEM_AFTER_DOUBLE_SPENT_DELAY
from .lnsweep import KeepWatchingTXO, SweepInfo

if TYPE_CHECKING:
    from .network import Network
    from .lnworker import LNWallet
    from .lnchannel import AbstractChannel


class LNWatcher(Logger, EventListener):
    MAX_CALLBACK_TRIGGER_DELAY_SEC = 600
    CALLBACK_LOOP_POLL_INTERVAL_SEC = 5

    def __init__(self, lnworker: 'LNWallet'):
        self.lnworker = lnworker
        Logger.__init__(self)
        self.adb = lnworker.wallet.adb
        self.config = lnworker.config
        self.callbacks = {}  # type: Dict[str, Callable[[], Awaitable[None]]]  # address -> lambda function
        self.network = None
        self.register_callbacks()
        self._pending_force_closes = set()
        self.taskgroup = OldTaskGroup()
        self._last_callback_trigger_ts = 0

    def start_network(self, network: 'Network'):
        assert not self.network, "already started?"
        self.network = network
        asyncio.run_coroutine_threadsafe(self._main_loop(), util.get_asyncio_loop())

    async def stop(self):
        await self.taskgroup.cancel_remaining()
        self.unregister_callbacks()

    async def _main_loop(self):
        self.logger.debug("starting taskgroup")
        try:
            async with self.taskgroup as group:
                await group.spawn(self._callback_loop())  # keeps group alive
        except Exception:
            self.logger.exception("taskgroup crashed")
        finally:
            self.logger.debug("taskgroup stopped")

    async def _callback_loop(self):
        """
        Triggers the callbacks if no event has triggered them within the
        last MAX_CALLBACK_TRIGGER_DELAY_SEC (e.g. during a prolonged time without new blocks)
        """
        ts_start = now()
        while True:
            max_delay = self.MAX_CALLBACK_TRIGGER_DELAY_SEC
            if now() - ts_start < max_delay:
                max_delay /= 10  # if wallet just recently opened, be much more eager
            time_since_last_cb_trigger = now() - self._last_callback_trigger_ts
            if time_since_last_cb_trigger > max_delay:
                await self.trigger_callbacks()
            await asyncio.sleep(self.CALLBACK_LOOP_POLL_INTERVAL_SEC)

    def remove_callback(self, address: str) -> None:
        self.callbacks.pop(address, None)

    def add_callback(
        self,
        address: str,
        callback: Callable[[], Awaitable[None]],
        *,
        subscribe: bool = True,
    ) -> None:
        if subscribe:
            # FIXME even when called with subscribe=False, adb likely already has this address.
            #   wallet.adb==lnwatcher.adb, and adb.db==wallet.db, which is persisted to disk.
            #   A call to adb.add_address at any time will add the address to the persistent DB.
            #   So in practice adb has the channel-related and swap-related addresses we *ever*
            #   subscribed to, and will have adb.synchronizer sub to them again.
            #   (even for old redeemed channels and old swaps)
            self.adb.add_address(address)
        self.callbacks[address] = callback

    async def trigger_callbacks(self, *, requires_synchronizer: bool = True):
        if requires_synchronizer and not self.adb.synchronizer:
            self.logger.debug("synchronizer not set yet")
            return
        for address, callback in list(self.callbacks.items()):
            try:
                await callback()
            except Exception:
                self.logger.exception(f"LNWatcher callback failed {address=}")
        # send callback to GUI
        util.trigger_callback('wallet_updated', self.lnworker.wallet)
        self._last_callback_trigger_ts = now()

    @event_listener
    async def on_event_blockchain_updated(self, *args):
        await self.trigger_callbacks()

    @event_listener
    async def on_event_adb_added_tx(self, adb, tx_hash, tx):
        # called if we add local tx
        if adb != self.adb:
            return
        await self.trigger_callbacks()

    @event_listener
    async def on_event_adb_added_verified_tx(self, adb, tx_hash):
        if adb != self.adb:
            return
        await self.trigger_callbacks()

    @event_listener
    async def on_event_adb_set_up_to_date(self, adb):
        if adb != self.adb:
            return
        await self.trigger_callbacks()

    def add_channel(self, chan: 'AbstractChannel') -> None:
        outpoint = chan.funding_outpoint.to_str()
        address = chan.get_funding_address()
        callback = lambda: self.check_onchain_situation(address, outpoint)
        self.add_callback(address, callback, subscribe=chan.need_to_subscribe())

    @ignore_exceptions
    @log_exceptions
    async def check_onchain_situation(self, address: str, funding_outpoint: str) -> None:
        # early return if address has not been added yet
        if not self.adb.is_mine(address):
            return
        # inspect_tx_candidate might have added new addresses, in which case we return early
        # note: maybe we should wait until adb.is_up_to_date... (?)
        funding_txid = funding_outpoint.split(':')[0]
        funding_height = self.adb.get_tx_height(funding_txid)
        closing_txid = self.adb.get_spender(funding_outpoint)
        closing_height = self.adb.get_tx_height(closing_txid)
        if closing_txid:
            self.adb.subscribe_to_outputs(closing_txid)
            closing_tx = self.adb.get_transaction(closing_txid)
            if closing_tx:
                keep_watching = await self.sweep_commitment_transaction(funding_outpoint, closing_tx)
                if not keep_watching:
                    self.remove_callback(address)
            else:
                self.logger.info(f"channel {funding_outpoint} closed by {closing_txid}. still waiting for tx itself...")
                keep_watching = True
        else:
            keep_watching = True
        await self.update_channel_state(
            funding_outpoint=funding_outpoint,
            funding_txid=funding_txid,
            funding_height=funding_height,
            closing_txid=closing_txid,
            closing_height=closing_height,
            keep_watching=keep_watching)

    def diagnostic_name(self):
        return f"{self.lnworker.wallet.diagnostic_name()}-LNW"

    async def update_channel_state(
            self, *, funding_outpoint: str, funding_txid: str,
            funding_height: TxMinedInfo, closing_txid: str,
            closing_height: TxMinedInfo, keep_watching: bool) -> None:
        chan = self.lnworker.channel_by_txo(funding_outpoint)
        if not chan:
            return
        chan.update_onchain_state(
            funding_txid=funding_txid,
            funding_height=funding_height,
            closing_txid=closing_txid,
            closing_height=closing_height,
            keep_watching=keep_watching)
        if closing_height.conf > 0:
            self._pending_force_closes.discard(chan)
        await self.lnworker.handle_onchain_state(chan)

    async def sweep_commitment_transaction(self, funding_outpoint: str, closing_tx: Transaction) -> bool:
        """This function is called when a channel was closed. In this case
        we need to check for redeemable outputs of the commitment transaction
        or spenders down the line (HTLC-timeout/success transactions).

        Returns whether we should continue to monitor.

        Side-effects:
          - sets defaults labels
          - populates wallet._accounting_addresses
        """
        assert closing_tx
        chan = self.lnworker.channel_by_txo(funding_outpoint)
        if not chan:
            return False
        local_height = self.adb.get_local_height()
        # detect who closed and get information about how to claim outputs
        is_local_ctx, sweep_info_dict = chan.get_ctx_sweep_info(closing_tx)
        # note: we need to keep watching *at least* until the closing tx is deeply mined,
        #       possibly longer if there are TXOs to sweep
        keep_watching = not self.adb.is_deeply_mined(closing_tx.txid())
        # create and broadcast transactions
        for prevout, sweep_info in sweep_info_dict.items():
            prev_txid, prev_index = prevout.split(':')
            name = sweep_info.name + ' ' + chan.get_id_for_log()
            self.lnworker.wallet.set_default_label(prevout, name)
            if isinstance(sweep_info, KeepWatchingTXO):  # haven't yet decided if we want to sweep
                keep_watching |= sweep_info.until_height > local_height
                continue
            assert isinstance(sweep_info, SweepInfo), sweep_info
            if not self.adb.get_transaction(prev_txid):
                # do not keep watching if prevout does not exist
                self.logger.info(f'prevout does not exist for {name}: {prevout}')
                continue
            watch_sweep_info = self.maybe_redeem(sweep_info)
            spender_txid = self.adb.get_spender(prevout)  # note: LOCAL spenders don't count
            spender_tx = self.adb.get_transaction(spender_txid) if spender_txid else None
            if spender_tx:
                # the spender might be the remote, revoked or not
                htlc_sweepinfo = chan.maybe_sweep_htlcs(closing_tx, spender_tx)
                if htlc_sweepinfo:
                    self.adb.subscribe_to_outputs(spender_txid)
                for prevout2, htlc_sweep_info in htlc_sweepinfo.items():
                    self.lnworker.wallet.set_default_label(prevout2, htlc_sweep_info.name)
                    if isinstance(htlc_sweep_info, KeepWatchingTXO):  # haven't yet decided if we want to sweep
                        keep_watching |= htlc_sweep_info.until_height > local_height
                        continue
                    assert isinstance(htlc_sweep_info, SweepInfo), htlc_sweep_info
                    watch_htlc_sweep_info = self.maybe_redeem(htlc_sweep_info)
                    htlc_tx_spender = self.adb.get_spender(prevout2)
                    if htlc_tx_spender:
                        keep_watching |= not self.adb.is_deeply_mined(htlc_tx_spender)
                        self.maybe_add_accounting_address(htlc_tx_spender, htlc_sweep_info)
                    else:
                        keep_watching |= watch_htlc_sweep_info
                keep_watching |= not self.adb.is_deeply_mined(spender_txid)
                self.maybe_extract_preimage(chan, spender_tx, prevout)
                self.maybe_add_accounting_address(spender_txid, sweep_info)
            else:
                keep_watching |= watch_sweep_info
            self.maybe_add_pending_forceclose(
                chan=chan,
                spender_txid=spender_txid,
                is_local_ctx=is_local_ctx,
                sweep_info=sweep_info,
            )
        return keep_watching

    def get_pending_force_closes(self):
        return self._pending_force_closes

    def maybe_redeem(self, sweep_info: 'SweepInfo') -> bool:
        """ returns 'keep_watching' """
        try:
            self.lnworker.wallet.txbatcher.add_sweep_input('lnwatcher', sweep_info)
        except BelowDustLimit:
            self.logger.debug(f"maybe_redeem: BelowDustLimit: {sweep_info.name}")
            # utxo is considered dust at *current* fee estimates.
            # but maybe the fees atm are very high? We will retry later.
            pass
        except NoDynamicFeeEstimates:
            self.logger.debug(f"maybe_redeem: NoDynamicFeeEstimates: {sweep_info.name}")
            pass  # will retry later
        if sweep_info.is_anchor():
            return False
        return True

    def maybe_extract_preimage(self, chan: 'AbstractChannel', spender_tx: Transaction, prevout: str):
        if not spender_tx.is_complete():
            self.logger.info('spender tx is unsigned')
            return
        txin_idx = spender_tx.get_input_idx_that_spent_prevout(TxOutpoint.from_str(prevout))
        assert txin_idx is not None
        spender_txin = spender_tx.inputs()[txin_idx]
        chan.extract_preimage_from_htlc_txin(
            spender_txin,
            is_deeply_mined=self.adb.is_deeply_mined(spender_tx.txid()),
        )

    def maybe_add_accounting_address(self, spender_txid: str, sweep_info: 'SweepInfo'):
        spender_tx = self.adb.get_transaction(spender_txid) if spender_txid else None
        if not spender_tx:
            return
        for i, txin in enumerate(spender_tx.inputs()):
            if txin.prevout == sweep_info.txin.prevout:
                break
        else:
            return
        if sweep_info.name in ['offered-htlc', 'received-htlc']:
            # always consider ours
            pass
        else:
            witness = txin.witness_elements()
            for sig in witness:
                # fixme: verify sig is ours
                witness2 = sweep_info.txin.make_witness(sig)
                if txin.witness == witness2:
                    break
            else:
                self.logger.info(f"signature not found {sweep_info.name}, {txin.prevout.to_str()}")
                return
        self.logger.info(f'adding txin address {sweep_info.name}, {txin.prevout.to_str()}')
        prev_txid, prev_index = txin.prevout.to_str().split(':')
        prev_tx = self.adb.get_transaction(prev_txid)
        txout = prev_tx.outputs()[int(prev_index)]
        self.lnworker.wallet._accounting_addresses.add(txout.address)

    def maybe_add_pending_forceclose(
        self,
        *,
        chan: 'AbstractChannel',
        spender_txid: Optional[str],
        is_local_ctx: bool,
        sweep_info: 'SweepInfo',
    ) -> None:
        """Adds chan into set of ongoing force-closures if the user should keep the wallet open, waiting for it.
        (we are waiting for ctx to be confirmed and there are received htlcs)
        """
        if is_local_ctx and sweep_info.name == 'received-htlc':
            cltv = sweep_info.cltv_abs
            assert cltv is not None, f"missing cltv for {sweep_info}"
            if self.adb.get_local_height() > cltv + REDEEM_AFTER_DOUBLE_SPENT_DELAY:
                # We had plenty of time to sweep. The remote also had time to time out the htlc.
                # Maybe its value has been ~dust at current and past fee levels (every time we checked).
                # We should not keep warning the user forever.
                return
            tx_mined_status = self.adb.get_tx_height(spender_txid)
            if tx_mined_status.height() == TX_HEIGHT_LOCAL:
                self._pending_force_closes.add(chan)
